using System;
using System.Collections.Generic;
using System.Linq;

using Syntriax.Engine.Core;
using Syntriax.Engine.Core.Serialization;

using YamlDotNet.Core;
using YamlDotNet.Core.Events;
using YamlDotNet.Serialization;

namespace Syntriax.Engine.Serializers.Yaml;

public class UniverseConverter : EngineTypeYamlSerializerBase<IUniverse>
{
    public override IUniverse? Read(IParser parser, Type type, ObjectDeserializer rootDeserializer)
    {
        bool isTrackingController = ProgressionTracker.Progression.ApproximatelyEquals(0f);
        ProgressionTracker.Set(isTrackingController ? 0.1f : ProgressionTracker.Progression, "Reading universe");

        string id;

        IUniverse universe;
        IStateEnable stateEnable;
        List<IUniverseObject> universeObjects;

        parser.Consume<MappingStart>();

        if (parser.Consume<Scalar>().Value.CompareTo(nameof(IUniverse.Id)) != 0)
            throw new();
        id = parser.Consume<Scalar>().Value;

        if (parser.Consume<Scalar>().Value.CompareTo(SERIALIZED_SCALAR_NAME) != 0)
            throw new();
        SerializedClass instanceSerializedClass = (SerializedClass)rootDeserializer(typeof(SerializedClass))!;
        ProgressionTracker.Set(isTrackingController ? .2f : ProgressionTracker.Progression, $"Creating {instanceSerializedClass.Type}");
        universe = (IUniverse)instanceSerializedClass.CreateInstance(EntityRegistry);

        if (parser.Consume<Scalar>().Value.CompareTo(nameof(IUniverse.StateEnable)) != 0)
            throw new();
        stateEnable = (IStateEnable)rootDeserializer(typeof(IStateEnable))!;

        ProgressionTracker.Set(isTrackingController ? .5f : ProgressionTracker.Progression, $"Reading universe objects");
        if (parser.Consume<Scalar>().Value.CompareTo(nameof(IUniverse.UniverseObjects)) != 0)
            throw new();
        universeObjects = (List<IUniverseObject>)rootDeserializer(typeof(List<IUniverseObject>))!;

        parser.Consume<MappingEnd>();

        universe.Id = id;

        stateEnable.Assign(universe);
        universe.Assign(stateEnable);

        ProgressionTracker.Set(isTrackingController ? .9f : ProgressionTracker.Progression, "Registering universe objects");
        for (int i = 0; i < universeObjects.Count; i++)
        {
            IUniverseObject uo = universeObjects[i];
            ProgressionTracker.Set(isTrackingController ? .9f + .1f * ((float)i / universeObjects.Count) : ProgressionTracker.Progression, $"Registering {uo.Name}");

            universe.Register(uo);
        }

        ProgressionTracker.Set(isTrackingController ? 1f : ProgressionTracker.Progression, $"Created {instanceSerializedClass.Type}");
        return universe;
    }

    public override void WriteYaml(IEmitter emitter, object? value, Type type, ObjectSerializer serializer)
    {
        IUniverse universe = (IUniverse)value!;

        bool isTrackingController = ProgressionTracker.Progression.ApproximatelyEquals(0f);
        ProgressionTracker.Set(isTrackingController ? .25f : ProgressionTracker.Progression, $"Serializing universe");

        IEnumerable<IUniverseObject> rootUniverseObjects = universe.UniverseObjects.Where(uo => uo.Parent is null);

        emitter.Emit(new MappingStart());

        emitter.Emit(new Scalar(nameof(IUniverse.Id)));
        emitter.Emit(new Scalar(universe.Id));

        emitter.Emit(new Scalar(SERIALIZED_SCALAR_NAME));
        serializer(new SerializedClass(universe));

        emitter.Emit(new Scalar(nameof(IUniverse.StateEnable)));
        serializer(universe.StateEnable);

        emitter.Emit(new Scalar(nameof(IUniverse.UniverseObjects)));
        serializer(rootUniverseObjects);

        ProgressionTracker.Set(isTrackingController ? 1f : ProgressionTracker.Progression, $"Serialized universe");
        emitter.Emit(new MappingEnd());
    }
}
